! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
!
! Copyright - 1996 Intel Corporation
!
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
!
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
!
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
!
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
!
! ----------------------------------------------------------------------
! **********************************************************************
*                          Functional test for MPI_Recv
*
*  This test verifies MPI_Recv sorts based on the communicator
*  specified.  3 communicators are created, and an assortment of messages
*  are sent from rank 0 to rank 1, using the first 2 communicators.
*
*  Rank 1 starts a receive on comm 3 (which should not be satisfied until
*  the end), then the receives specified above.  Messages should be 
*  received in the order sent for each communicator, not in the order
*  the receives were posted (which is different).
! **********************************************************************
#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'

      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


      INTEGER    NUMMESG
      PARAMETER (NUMMESG = 20)
!                               number of messages to Xsend/Recv

      INTEGER    FAIL
      INTEGER    LOOP_CNT
!                               counts total number of failures, loops
      INTEGER    IERR
      INTEGER    ERR
      INTEGER    SIZE
!                               return values from MPI calls
      INTEGER    CNT_LEN
!                               length of received data
      LOGICAL    FLAG
!                               returned from Test

      INTEGER    COMM(64)
!                               additional communicators to sort on
!                               0 = MPI_COMM_WORLD, 2-3 are DUPed
!                               No messages will be sent on comm3
!                               until all others have been received.

      INTEGER    I
      INTEGER    J
      INTEGER    ERRORS

      CHARACTER*(IOSIZE)  INFOBUF
      CHARACTER*32   TESTNAME
      CHARACTER*(MPI_MAX_ERROR_STRING) ERRORSTRING

      CHARACTER*20       SENDBUF(3*NUMMESG)
      CHARACTER*20       RECVBUF(3*NUMMESG+1)
      INTEGER    NUMELM
      PARAMETER (NUMELM = 20)
!                               number of elements to send/receive
!                               MUST MATCH CHARACTER LENGTH ABOVE!


      CHARACTER*(NUMMESG * (8*NUMELM+MPI_BSEND_OVERHEAD+100)) BSEND_BUFF

      INTEGER   RECV_STAT(MPI_STATUS_SIZE)
      INTEGER   SEND_STAT(MPI_STATUS_SIZE)

      INTEGER   RECV_REQ

 90   FORMAT(I2, A13, I2)
 97   FORMAT(A,INT_FMT,A,A)
 98   FORMAT(A,INT_FMT,A,INT_FMT)
 99   FORMAT(A,INT_FMT)
 189  FORMAT(A,INT_FMT,A)

!
!     Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

!
!     Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
      TESTNAME = 'MPI_Recv_comm'
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF


!
!     Initialize the loop/error counters
!
      LOOP_CNT = 0
      FAIL = 0

!
!     Need at least 2 nodes.
!
      IF (MPITEST_NUMP .LT. 2) THEN
        INFOBUF = 'At least 2 ranks required to run this test'
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

!
!     Create some communicators.  Use the first 3 to send some messages.
!     Leave the last one idle until the end.
!
      COMM(1) = MPI_COMM_WORLD

      DO 50 I = 2, 64
        CALL MPI_COMM_DUP(MPI_COMM_WORLD, COMM(I), IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          INFOBUF=' '
          WRITE(INFOBUF,99) 'MPI_COMM_DUP() returned', IERR
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          INFOBUF=' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
          FAIL = FAIL + 1
        END IF
 50   CONTINUE

      CALL MPI_BARRIER(COMM(1), IERR)

      IF (MPITEST_ME .LT. 2) THEN

!       Initialize Send Buffers
        DO 100 I=1, NUMMESG*3
          WRITE(SENDBUF(I),90) I, ' this buffer ', I
 100    CONTINUE

        IF (MPITEST_ME .EQ. 0) THEN
!
!         Sender
!
          CALL MPI_BUFFER_ATTACH(BSEND_BUFF,  &
              NUMMESG * (8*NUMELM + MPI_BSEND_OVERHEAD+100), IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            INFOBUF=' '
            WRITE(INFOBUF,99) 'MPI_BUFFER_ATTACH() returned', IERR
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            INFOBUF=' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
            FAIL = FAIL + 1
          END IF

          J = 0
          DO 200 I=1, NUMMESG
            J = J + 1
            IF (J .EQ. 3) J = 1
            LOOP_CNT = LOOP_CNT + 1
            CALL MPI_SEND(SENDBUF(3*I-2), NUMELM, MPI_CHARACTER, 1, &
                 1, COMM(J), IERR)
            IF (IERR .NE. MPI_SUCCESS) THEN
              INFOBUF=' '
              WRITE(INFOBUF,99) 'MPI_SEND() returned', IERR
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              INFOBUF=' '
              CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
              FAIL = FAIL + 1
            END IF


            J = J + 1
            IF (J .EQ. 3) J = 1
            LOOP_CNT = LOOP_CNT + 1
            CALL MPI_BSEND(SENDBUF(3*I-1), NUMELM, MPI_CHARACTER, &
                  1, 1, COMM(J), IERR)
            IF (IERR .NE. MPI_SUCCESS) THEN
              INFOBUF=' '
              WRITE(INFOBUF,99) 'MPI_BSEND() returned', IERR
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              INFOBUF=' '
              CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
              FAIL = FAIL + 1
            END IF


            J = J + 1
            IF (J .EQ. 3) J = 1
            LOOP_CNT = LOOP_CNT + 1
            CALL MPI_SSEND(SENDBUF(3*I), NUMELM, MPI_CHARACTER, &
                  1, 1, COMM(J), IERR)
            IF (IERR .NE. MPI_SUCCESS) THEN
              INFOBUF=' '
              WRITE(INFOBUF,99) 'MPI_SSEND() returned', IERR
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              INFOBUF=' '
              CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
              FAIL = FAIL + 1
            END IF

 200      CONTINUE

!
!         Wait before sending the final message on the last comm.
!
          CALL MPI_BARRIER(COMM(3), IERR)
          CALL MPI_SEND(SENDBUF(1), NUMELM, MPI_CHARACTER, 1, &
                 1, COMM(3), IERR)
          LOOP_CNT = LOOP_CNT + 1
          IF (IERR .NE. MPI_SUCCESS) THEN
            INFOBUF=' '
            WRITE(INFOBUF,99) 'MPI_SEND returned', IERR
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            INFOBUF=' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
            FAIL = FAIL + 1
          END IF

          CALL MPI_BUFFER_DETACH(BSEND_BUFF, SIZE, IERR) 


        ELSE
!
!         Receiver
!
          LOOP_CNT = 1 + LOOP_CNT
          CALL MPI_IRECV(RECVBUF(3*NUMMESG+1), NUMELM, &
              MPI_CHARACTER, 0, 1, COMM(3), RECV_REQ, IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            WRITE(INFOBUF,99) 'MPI_IRECV() returned', IERR
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            INFOBUF=' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
          END IF

          J = 0
          DO 300 I=1, 3*NUMMESG
            J = J + 1
            IF (J .EQ. 3) J = 1
            RECVBUF(I)='                    '

            LOOP_CNT = 1 + LOOP_CNT
            CALL MPI_RECV(RECVBUF(I), NUMELM, MPI_CHARACTER, 0, 1, &
               COMM(J), RECV_STAT, IERR)
            IF (IERR .NE. MPI_SUCCESS) THEN
              WRITE(INFOBUF,99) 'MPI_IRECV() returned', IERR
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              INFOBUF=' '
              CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
              CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
            END IF
 300      CONTINUE

          DO 400 I=1, 3*NUMMESG
            IF (SENDBUF(I)(1:20) .NE. RECVBUF(I)(1:20)) THEN
              INFOBUF=' '
              WRITE(INFOBUF,97) &
               'Unexpected data in recv', I, ' = ' , &
               RECVBUF(I)
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              FAIL = FAIL + 1
            END IF

            CALL MPI_GET_COUNT(RECV_STAT, MPI_CHARACTER, CNT_LEN, &
                IERR)
            IF (IERR .NE. MPI_SUCCESS) THEN
              WRITE(INFOBUF,99) 'MPI_GET_COUNT() returned', IERR
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              INFOBUF=' '
              CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
              FAIL = FAIL + 1
            END IF

            IF (NUMELM .NE. CNT_LEN) THEN
              INFOBUF=' '
              WRITE(INFOBUF,98) &
               'Incorrect length in recv ', I, ' = ' , CNT_LEN
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              FAIL = FAIL + 1
            END IF

            IF (RECV_STAT(MPI_TAG) .NE. 1) THEN
              INFOBUF=' '
              WRITE(INFOBUF,98) &
               'Incorrect tag in recv ',I,' = ', RECV_STAT(MPI_TAG)
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              FAIL = FAIL + 1
            END IF

 400      CONTINUE

!
!         Make sure we haven't received anything on COMM(3).
!
          CALL MPI_TEST(RECV_REQ, FLAG, RECV_STAT, &
              IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            WRITE(INFOBUF,99) 'MPI_TEST() returned', IERR
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            INFOBUF=' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
          END IF
          IF (FLAG .NEQV. .FALSE.) THEN
            INFOBUF='Received Message on COMM3, but none were sent!'
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
            FAIL = FAIL + 1
          END IF

!
!         OK, receive the last message.
!
          CALL MPI_BARRIER(COMM(3), IERR)
          CALL MPI_WAIT(RECV_REQ, RECV_STAT, IERR)


        END IF
      ELSE
!
!       Ranks >= 2 need to match BARRIERs above
!
        CALL MPI_BARRIER(COMM(3), IERR)

      END IF

      DO 550 I = 2, 3
        CALL MPI_COMM_FREE(COMM(I), IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          INFOBUF=' '
          WRITE(INFOBUF,99) 'MPI_COMM_FREE() returned', IERR
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          INFOBUF=' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
          FAIL = FAIL + 1
        END IF
 550  CONTINUE

!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_CNT - FAIL, FAIL, 0, TESTNAME)

      CALL MPI_FINALIZE(IERR)

      END

