! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
! **********************************************************************
!                    Error test for MPI_SCAN()
!
! This test verifies that the correct error is returned if MPI_SCAN()
! is called with an invalid argument.
!
! 1)  Invalid type/operator pairs.
! 2)  Use mis-matched types.
! 3)  Use invalid datatype.
! 4)  Use invalid size (-1).
! 5)  Use MPI_COMM_NULL.
! 6)  Use a freed communicator.
! 7)  Use an inter-communicator.
!
! In all cases, expect to receive appropriate error.
!
! Test history:
!    1  02/12/96     gt   Created
!
! **********************************************************************
#include "f90options.h"


      SUBROUTINE TEST_OP(TYPE, OP, PASS, FAIL, VERIFY)
! **********************************************************************
!     Test invalid datatype/operation pairs.
!
! **********************************************************************
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


      INTEGER TYPE
      INTEGER OP
      INTEGER PASS
      INTEGER FAIL
      INTEGER VERIFY
      INTEGER INPUT(MAX_RANKS)
      INTEGER OUTPUT(MAX_RANKS)
      INTEGER SIZE
      INTEGER IERR
      INTEGER ERR
      INTEGER ERRORCLASS

      CHARACTER*(IOSIZE)  INFOBUF
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING

 100  FORMAT(A200)
 198  FORMAT(A,INT_FMT,A,INT_FMT,A)
 199  FORMAT(A,INT_FMT,A,INT_FMT,A,INT_FMT,A)

      IERR = 0
      CALL MPI_SCAN(INPUT, OUTPUT, 1, TYPE, OP, MPI_COMM_WORLD, IERR)

      IF (IERR .EQ. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 198) 'MPI_SCAN() w/ type ', TYPE, ' op ', &
            OP, ' did not FAIL'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
      ELSE
        CALL MPI_ERROR_CLASS(IERR, ERRORCLASS, ERR)
        IF (ERRORCLASS .NE. MPI_ERR_TYPE .AND.  &
            ERRORCLASS .NE. MPI_ERR_OP) THEN 
          FAIL = FAIL + 1
          WRITE(INFOBUF, 199) 'MPI_SCAN() w/ type ', TYPE, ' op ', &
            OP, ' returned ', ERRORCLASS, ' expected MPI_ERR_TYPE'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING=' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF, 100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ELSE
          PASS = PASS + 1
        END IF
      END IF


      END

! ***********************************************************************

      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


      EXTERNAL TEST_OP

!                           counts total number of PASSes/FAILures/VERIFY
      INTEGER PASS
      INTEGER FAIL              
      INTEGER VERIFY
!                           return value from MPI calls
      INTEGER IERR
      INTEGER ERR
!                           error class of IERR
      INTEGER ERRORCLASS
!                           used to split a communicator
      INTEGER COLOR
      INTEGER SIZE 

      CHARACTER*(IOSIZE)  INFOBUF
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
      CHARACTER*32   TESTNAME


!                           MPI communicators
      INTEGER COMM

!                           Input to SCAN
      INTEGER INPUT(MAX_RANKS)
      INTEGER OUTPUT(MAX_RANKS)

 98   FORMAT(A,INT_FMT,A)
 99   FORMAT(A,INT_FMT)
 100  FORMAT(A200)


!
!  Initialize the MPI environment and test environment.
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR 
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

      TESTNAME='MPI_Scan_err1'
      CALL MPITEST_INIT(IERR)
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF

      PASS = 0
      FAIL = 0
      VERIFY = 0
! 
!  Set an errorhandler so we get control back.  
! 
      CALL MPI_ERRHANDLER_SET(MPI_COMM_WORLD, MPI_ERRORS_RETURN, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 99) ' MPI_ERRORHANDLER_SET returned ', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

!
!  Invalid type for operation
!
      CALL TEST_OP(MPI_BYTE, MPI_MAX, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_BYTE, MPI_MAXLOC, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_BYTE, MPI_MIN, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_BYTE, MPI_MINLOC, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_BYTE, MPI_SUM, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_BYTE, MPI_PROD, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_BYTE, MPI_LAND, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_BYTE, MPI_LOR, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_BYTE, MPI_LXOR, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_COMPLEX, MPI_MAX, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_COMPLEX, MPI_MAXLOC, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_COMPLEX, MPI_MIN, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_COMPLEX, MPI_MINLOC, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_COMPLEX, MPI_LAND, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_COMPLEX, MPI_LOR, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_COMPLEX, MPI_LXOR, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_COMPLEX, MPI_BAND, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_COMPLEX, MPI_BOR, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_COMPLEX, MPI_BXOR, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_LOGICAL, MPI_MAX, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_LOGICAL, MPI_MAXLOC, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_LOGICAL, MPI_MIN, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_LOGICAL, MPI_MINLOC, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_LOGICAL, MPI_PROD, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_LOGICAL, MPI_SUM, PASS, FAIL, VERIFY)
c      CALL TEST_OP(MPI_LOGICAL, MPI_BAND, PASS, FAIL, VERIFY)
c      CALL TEST_OP(MPI_LOGICAL, MPI_BOR, PASS, FAIL, VERIFY)
c      CALL TEST_OP(MPI_LOGICAL, MPI_BXOR, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_INTEGER, MPI_MAXLOC, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_INTEGER, MPI_MINLOC, PASS, FAIL, VERIFY)
c      CALL TEST_OP(MPI_INTEGER, MPI_LAND, PASS, FAIL, VERIFY)
c      CALL TEST_OP(MPI_INTEGER, MPI_LOR, PASS, FAIL, VERIFY)
c      CALL TEST_OP(MPI_INTEGER, MPI_LXOR, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_REAL, MPI_MAXLOC, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_REAL, MPI_MINLOC, PASS, FAIL, VERIFY)
c      CALL TEST_OP(MPI_REAL, MPI_LAND, PASS, FAIL, VERIFY)
c      CALL TEST_OP(MPI_REAL, MPI_LOR, PASS, FAIL, VERIFY)
c      CALL TEST_OP(MPI_REAL, MPI_LXOR, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_REAL, MPI_BAND, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_REAL, MPI_BOR, PASS, FAIL, VERIFY)
      CALL TEST_OP(MPI_REAL, MPI_BXOR, PASS, FAIL, VERIFY)

!
!  Report overall results
!
      CALL MPITEST_REPORT(PASS, FAIL, VERIFY, TESTNAME)

      CALL MPI_FINALIZE(IERR)

      END
