! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
! **********************************************************************
!                          Test of MPI_Scan_user()
! 
! This code tests the functionality of the MPI_Scan() funtion for user-defined
! operations.
! 
! This function acts much like a Reduce_scatter(), except the reduced data which
! arrives at process "i" has only been reduced with data from processes
! 0, 1, 2, ..., i.
! 
! This test initializes the ith processes send data with "i".  Then the Scan is
! performed, then the data is checked.
! 
! The operations to be looped over are in the array MPITEST_default_ops[].
! This array must be initialized at runtime, after the call to MPI_Init().
! **********************************************************************

#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'

      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'

      INCLUDE 'externalf90.h'
      INCLUDE 'reduce.h'

      INTEGER MPITEST_DEFAULT_OPS(10)

      INTEGER    COMM_INDEX
!                               the array index of the current comm
      INTEGER    COMM_TYPE
!                               the index of the current communicator type
      INTEGER    COMM_COUNT
!                               loop counter for communicator loop
      INTEGER    COMM_SIZE
!                               number of ranks in the current communicator
      LOGICAL    INTER_FLAG
!                               used to test if intercommunicator
      INTEGER    TYPE_COUNT
!                               number of data types to test
      INTEGER    TEST_TYPE
!                               index of current data type
      INTEGER    LENGTH_COUNT
!                               number of data lengths to test
      INTEGER    MAX_LENGTH
!                               max message length to test
      INTEGER    MAX_BYTE_LENGTH
!                               max byte length to test
      INTEGER    MAX_BUFF_LENGTH
!                               max length that will fit in buffers
      INTEGER    BYTE_LENGTH
!                               current byte length to test
      INTEGER    LENGTH
!                               current length to test 
      INTEGER    NUM_OPS
!                               number of operations to test
      INTEGER    OP_COUNT
!                               index of current operation to test

      INTEGER    FAIL
      INTEGER    LOOP_CNT
!                               counts total number of failures, loops
      INTEGER    ERROR, ERROR2
!                               number of data errors found in buffer
      INTEGER    IERR
      INTEGER    ERR
!                               return value from MPI calls
      INTEGER    COMM
!                               MPI communicator under test
      INTEGER    I, J, K, L

      LOGICAL    HASOP

      MPITEST_BUF_TYPE  SEND_BUFFER(MAX_BUFF_SIZE+MAX_RANKS)
      MPITEST_BUF_TYPE  RECV_BUFFER(MAX_BUFF_SIZE+MAX_RANKS)
      INTEGER    ERRSIZE

      MPITEST_BUF_TYPE VALUE

      CHARACTER*(IOSIZE)  INFOBUF
      CHARACTER*32   TESTNAME
       CHARACTER*(MPI_MAX_ERROR_STRING) ERRORSTRING

 99   FORMAT(A,INT_FMT)
 179  FORMAT(A,I6,A,I6,A,I6,A,I3,A,INT_FMT,A,I3)
 189  FORMAT(INT_FMT,A,A,I6,A,I6,A,I6,A,I3,A,INT_FMT,A,I3)

!
!     Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!     Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
      TESTNAME = 'MPI_Scan_user()'
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF


!
!     Initialize the loop/error counters
!
      LOOP_CNT = 0
      FAIL = 0

!
!     Get number of communicators, data types and message lengths to test
!
      CALL MPITEST_NUM_COMM_SIZES(COMM_COUNT)
      CALL MPITEST_NUM_DATATYPES(TYPE_COUNT)
      CALL MPITEST_NUM_MESSAGE_LENGTHS(LENGTH_COUNT)

!
!     Get max message length for test
!
      CALL MPITEST_GET_MAX_MESSAGE_LENGTH(MAX_BYTE_LENGTH)

!
!     Initialize operations
!
      CALL SET_DEFAULT_OPS(MPITEST_DEFAULT_OPS, NUM_OPS)

!
!     Loop through communicators
!
      DO 100 I=1,COMM_COUNT
        CALL MPITEST_GET_COMM_INDEX(I, COMM_INDEX)
        CALL MPITEST_GET_COMM_TYPE(I, COMM_TYPE)

        CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, COMM, &
              COMM_SIZE)

!       Skip to end of loop if this node is not part of current
!       communicator

        IF (COMM .EQ. MPI_COMM_NULL) GO TO 201

!       Skip if intercommunicator

        CALL MPI_COMM_TEST_INTER(COMM, INTER_FLAG, IERR)
        IF (INTER_FLAG .EQV. .TRUE.) GO TO 201

!
!       Loop through data types.
!
        DO 200 J=1,TYPE_COUNT

          CALL MPITEST_GET_DATATYPE(J, TEST_TYPE)
!
!         Convert the number of bytes in the maximum length message
!         into the number of elements of the current type 
          CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, MAX_BYTE_LENGTH, &
                MAX_LENGTH)
          CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, &
                MAX_BUFF_SIZE * MPITEST_BUF_EXTENT, MAX_BUFF_LENGTH)

          IF (MAX_LENGTH .GE. (MAX_BUFF_LENGTH - 1)) &
               MAX_LENGTH = MAX_BUFF_LENGTH - 1
          IF (MAX_LENGTH .LE. 0) MAX_LENGTH = 0


!
!         Loop through data lengths
!
          DO 300 K=1,LENGTH_COUNT
            CALL MPITEST_GET_MESSAGE_LENGTH(K, BYTE_LENGTH)
            CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, BYTE_LENGTH, &
                LENGTH)
            IF (LENGTH .GT. MAX_LENGTH) LENGTH = MAX_LENGTH

!
!           Loop through the different operations
!
            DO 350 OP_COUNT=1,NUM_OPS
              CALL HAS_OP(OP_COUNT, TEST_TYPE, HASOP)
              IF (HASOP .EQV. .FALSE.) &
                GO TO 350

!             Print an informational message
              IF (MPITEST_CURRENT_RANK .EQ. 0) THEN
                WRITE (INFOBUF,179) 'commindex',COMM_INDEX, &
                    ', commsize', COMM_SIZE,  &
                    ', commtype', COMM_TYPE, &
                    ', datatype', TEST_TYPE, ', length',  LENGTH, &
                    '. op', OP_COUNT
                CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
              END IF
              LOOP_CNT = LOOP_CNT + 1

!
!             Initialize send data, receive buffer
!
              CALL MPITEST_DATATEMPLATE_INIT(VALUE, &
                    MPITEST_CURRENT_RANK+1, TEST_TYPE)
              CALL MPITEST_INIT_BUFFER(TEST_TYPE, LENGTH, VALUE, &
                    SEND_BUFFER)
              CALL MPITEST_DATATEMPLATE_INIT(VALUE, &
                    -1, TEST_TYPE)
              CALL MPITEST_INIT_BUFFER(TEST_TYPE, LENGTH+1, VALUE, &
                    RECV_BUFFER)


              IERR = 0
              IF (LENGTH .NE. 0) &
                CALL MPI_SCAN(SEND_BUFFER, RECV_BUFFER, LENGTH, &
                   MPITEST_MPI_DATATYPES(TEST_TYPE), &
                   MPITEST_DEFAULT_OPS(OP_COUNT), COMM, IERR)
              IF (IERR .NE. MPI_SUCCESS) THEN
                WRITE(INFOBUF,99) 'MPI_SCAN() returned', IERR
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                INFOBUF=' '
                CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
                FAIL = FAIL + 1
              END IF
!
!               Check data for errors and overflow
!
              CALL GET_REDUCE_ANSWER(OP_COUNT, MPITEST_CURRENT_RANK + 1, &
                  VALUE, TEST_TYPE)
              CALL MPITEST_BUFFER_ERRORS(TEST_TYPE, LENGTH, VALUE, &
                  RECV_BUFFER, ERROR)
              CALL MPITEST_DATATEMPLATE_INIT(VALUE, &
                  -1, TEST_TYPE)
              CALL MPITEST_BUFFER_ERRORS_OV(TEST_TYPE, LENGTH, &
                  VALUE, RECV_BUFFER, ERROR2)
              IF ((ERROR + ERROR2) .NE. 0) THEN
                IF (IERR .EQ. MPI_SUCCESS) FAIL = FAIL + 1
                WRITE (INFOBUF,189) ERROR+ERROR2, &
                  ' errors in buffer, ', &
                  'commindex',COMM_INDEX,', commsize', COMM_SIZE, &
                  ', commtype', COMM_TYPE, &
                  ', datatype', TEST_TYPE, ', length',  LENGTH, &
                  '. op', OP_COUNT
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              END IF


 350        CONTINUE
      
 300      CONTINUE
      
 200    CONTINUE
 201    CONTINUE

        CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM, IERR)

 100  CONTINUE

!
!     Free the operations
!
      DO 600 I=1,NUM_OPS
        LOOP_CNT = LOOP_CNT + 1
        CALL MPI_OP_FREE(MPITEST_DEFAULT_OPS(I), IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,99) 'MPI_OP_FREE() returned', IERR
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          INFOBUF=' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
          FAIL = FAIL + 1
        END IF
 600  CONTINUE


!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_CNT - FAIL, FAIL, 0, TESTNAME)
      
      CALL MPI_FINALIZE(IERR)

      END  
  
  
