! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
!
! Copyright - 1996 Intel Corporation
!
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
!
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
!
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
!
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
!
! ----------------------------------------------------------------------
! **********************************************************************
!                           Test for MPI_TYPE_CONTIGUOUS()
!
! All rank will call MPI_TYPE_CONTIGUOUS() and then MPI_TYPE_COMMIT() to create
! a new user defined type using basic type defined in the types array in config.h.
! The root rank will then send a message using the new type to the receiving rank.
! The receiveing rank will do a recv.  The received buffer will be verified.
! Depending on whether MPITEST_STATUS_CHK is #defined or not, the output status
! object from MPI_RECV() may or may not be verified. 
!
! The input count to MPI_TYPE_CONTIGUOUS() is determined by each message length
! in config.h and the input oldtype to MPI_TYPE_CONTIGUOUS() is determined
! by the input types in config.h.
!
! This test may be run in any communicator with a minimum of 2 group members,
! with any data type, and with any non-negative message length.
!
! The MPITEST environment provides looping over communicator size,
! message length and the root's rank.  The properties of the loops are
! encoded in configuration arrays in the file config.h .
!
! MPI Calls dependencies for this test:
!   MPI_SEND(), MPI_RECV(), MPI_CANCEL(), MPI_INIT(), MPI_FINALIZE()
!   MPI_COMM_TEST_INTER(), MPI_ERROR_STRING(), MPI_TYPE_EXTENT(),
!   MPI_TYPE_CONTIGUOUS(), MPI_TYPE_COMMIT(), MPI_TYPE_FREE(),
!   [MPI_GET_COUNT(), MPI_ALLREDUCE(), MPI_COMM_RANK(), MPI_COMM_SIZE()]
!
! Test history:
!    1  07/29/96     simont       Original version
! ******************************************************************************/
!
#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'

      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


      INTEGER TEST_TYPE
      INTEGER LENGTH_COUNT
      INTEGER LENGTH
      INTEGER BYTE_LENGTH
      INTEGER MAX_LENGTH_COUNT
      INTEGER TEST_NUMP
      INTEGER COMM_INDEX
      INTEGER COMM_TYPE
      INTEGER TYPE_COUNT
      INTEGER MAX_TYPE_COUNT
      INTEGER COMM_COUNT
      INTEGER MAX_COMM_COUNT
      INTEGER ERROR, ERROR2
      INTEGER FAIL
      INTEGER SIZE
      INTEGER LOOP_CNT
      INTEGER IERR, IERR2
      INTEGER MAX_LENGTH
      INTEGER ROOT
      INTEGER TAG
      INTEGER DEST
!     WDG - 4/29/03 - The original code used MPITEST_AINT, which is 
!     incorrect for the Fortran binding
      INTEGER EXTENT

      INTEGER MPITEST_TAG 
      PARAMETER(MPITEST_TAG = 1) 

      INTEGER COMM, STATUS(MPI_STATUS_SIZE), NEWTYPE

      LOGICAL INTER_FLAG

#ifdef MPITEST_STATUS_CHK
      INTEGER COUNT
      INTEGER EXPECTED_COUNT
#endif

      MPITEST_BUF_TYPE VALUE
      MPITEST_BUF_TYPE BUFFER(MAX_BUFF_SIZE+1)

      CHARACTER*(IOSIZE)  INFOBUF
      CHARACTER*32        TESTNAME

      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING

 99   FORMAT(A, INT_FMT)
 100  FORMAT(A200)
 110  FORMAT(A)
 130  FORMAT(A, INT_FMT, A)
 140  FORMAT(A, INT_FMT, A, INT_FMT, A, INT_FMT)
 150  FORMAT(A, A, A, INT_FMT, A, INT_FMT)
 160  FORMAT(INT_FMT, A, INT_FMT, A, INT_FMT, A, INT_FMT, A, INT_FMT, &
             A, INT_FMT, A, INT_FMT, A, INT_FMT)

!
!  Initialize the MPI environment and test environment.
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
         WRITE(INFOBUF,99) 'MPI_INIT() returned ', IERR 
         CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      ENDIF

      TESTNAME='MPI_Type_contiguous_basic'
      CALL MPITEST_INIT(IERR)
      IF (MPITEST_ME .EQ. 0) THEN
         INFOBUF = 'Starting test '//TESTNAME
         CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      ENDIF

      FAIL = 0
      LOOP_CNT = 0

!     Find the maximum number of the communicators in config array
!
      CALL MPITEST_NUM_COMM_SIZES(MAX_COMM_COUNT)

      DO 300 COMM_COUNT = 1, MAX_COMM_COUNT
         CALL MPITEST_GET_COMM_INDEX(COMM_COUNT, COMM_INDEX)
         CALL MPITEST_GET_COMM_TYPE(COMM_COUNT, COMM_TYPE)

         CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, &
                                       COMM, TEST_NUMP)

         IF (COMM .EQ. MPI_COMM_NULL) GO TO 1000

         IF (TEST_NUMP .LT. 2) THEN
!           Skipping communicator with comm size < 2
!
            WRITE(INFOBUF, 130) &
               'Skipping communicator with comm_size < 2 (commtype: ', &
               COMM_TYPE, ') for this test'
            CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
         ELSE
            CALL MPI_COMM_TEST_INTER(COMM, INTER_FLAG, IERR)
            IF (IERR .NE. MPI_SUCCESS) THEN
               WRITE(INFOBUF,99) 'MPI_COMM_TEST_INTER() returned ', &
                  IERR 
               CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
               ERRORSTRING=' '
               CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, IERR2)
               WRITE(INFOBUF, 100) ERRORSTRING(1:SIZE)
               CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
            ENDIF

            IF (INTER_FLAG .EQV. .TRUE.) THEN
               WRITE(INFOBUF, 130) &
                  'Skipping inter communicator (commtype: ', &
                  COMM_TYPE, ') for this test'
               CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
            ELSE
               CALL MPITEST_NUM_DATATYPES(MAX_TYPE_COUNT)

               DO 400 TYPE_COUNT = 1, MAX_TYPE_COUNT, 1
                  CALL MPITEST_GET_DATATYPE(TYPE_COUNT, TEST_TYPE)

                  CALL MPI_TYPE_EXTENT( &
                     MPITEST_MPI_DATATYPES(TEST_TYPE), EXTENT, &
                        IERR)
                  IF (IERR .NE. MPI_SUCCESS) THEN
                     WRITE(INFOBUF,99)  &
                        'MPI_TYPE_EXTENT() returned ', &
                        IERR 
                     CALL MPITEST_MESSAGE(MPITEST_NONFATAL, &
                        INFOBUF)
                     ERRORSTRING=' '
                     CALL MPI_ERROR_STRING(IERR, ERRORSTRING, &
                        SIZE, IERR2)
                     WRITE(INFOBUF, 100) ERRORSTRING(1:SIZE)
                        CALL MPITEST_MESSAGE(MPITEST_FATAL, &
                        INFOBUF)
                  END IF

!                 Found out the number of message length in
!                 config array
!
                  CALL MPITEST_NUM_MESSAGE_LENGTHS(MAX_LENGTH_COUNT)

!                 Looping throught each length input
!
                  DO 500 LENGTH_COUNT = 1, MAX_LENGTH_COUNT, 1
                     CALL MPITEST_GET_MESSAGE_LENGTH(LENGTH_COUNT, &
                        LENGTH)

!                    Skipping length 0
!
                     IF (LENGTH .EQ. 0) THEN
                        IF (MPITEST_CURRENT_RANK .EQ. 0) THEN
                           WRITE(INFOBUF, 110) 'Skipping length = 0'
                           CALL MPITEST_MESSAGE(MPITEST_INFO1, &
                              INFOBUF)
                        END IF
                     ELSE IF (MPITEST_BUF_EXTENT * MAX_BUFF_SIZE .LT. &
                         EXTENT * LENGTH) THEN

!                       Make sure there is enough space in the buffer
!
                        WRITE(INFOBUF, 130) &
                        'Skipping too large length = ', LENGTH, &
                        ' for buffer space'
                     ELSE
!                       Create a user defined datatype 
!
                        CALL MPI_TYPE_CONTIGUOUS(LENGTH, &
                        MPITEST_MPI_DATATYPES(TEST_TYPE), NEWTYPE, &
                           IERR)
                        IF (IERR .NE. MPI_SUCCESS) THEN
                           WRITE(INFOBUF,99)  &
                              'MPI_TYPE_CONTIGUOUS() returned ', &
                              IERR 
                           CALL MPITEST_MESSAGE(MPITEST_NONFATAL, &
                              INFOBUF)
                           ERRORSTRING=' '
                           CALL MPI_ERROR_STRING(IERR, ERRORSTRING, &
                              SIZE, IERR2)
                           WRITE(INFOBUF, 100) ERRORSTRING(1:SIZE)
                           CALL MPITEST_MESSAGE(MPITEST_FATAL, &
                              INFOBUF)
                        END IF

                        CALL MPI_TYPE_COMMIT(NEWTYPE, IERR)
                        IF (IERR .NE. MPI_SUCCESS) THEN
                           WRITE(INFOBUF,99) &
                              'MPI_TYPE_COMMIT() returned ', IERR 
                           CALL MPITEST_MESSAGE(MPITEST_NONFATAL, &
                              INFOBUF)
                           ERRORSTRING=' '
                           CALL MPI_ERROR_STRING(IERR, &
                              ERRORSTRING, SIZE, IERR2)
                           WRITE(INFOBUF, 100) ERRORSTRING(1:SIZE)
                           CALL MPITEST_MESSAGE(MPITEST_FATAL, &
                              INFOBUF)
                        END IF

!                       Now test the newly created type in data transmission
!
                        DO 600 ROOT = 0, TEST_NUMP - 1, 1

!                          Print an informational message
!
                           IF (MPITEST_CURRENT_RANK .EQ. 0) THEN
                              WRITE(INFOBUF, 140) '(', LENGTH_COUNT, &
                              ', ', COMM_COUNT, ') length ', LENGTH
                              CALL MPITEST_MESSAGE(MPITEST_INFO1, &
                                 INFOBUF)
                           END IF

!                          Set up datatemplate for initialize buffer
!
                           CALL MPITEST_DATATEMPLATE_INIT(VALUE, &
                              MPITEST_CURRENT_RANK, TEST_TYPE)

!                          Initialize buffer
!
                           CALL MPITEST_INIT_BUFFER(TEST_TYPE, &
                              LENGTH + 1, VALUE, BUFFER)

                           LOOP_CNT = LOOP_CNT + 1

                           TAG = MPITEST_TAG

!                          Destination rank
!
                           DEST = ROOT + 1
                           IF (DEST .GE. TEST_NUMP) THEN
                              DEST = 0
                           END IF

                           IF (MPITEST_CURRENT_RANK .EQ. ROOT) THEN
                              WRITE(INFOBUF, 140) &
                                 'Sending from source: ', ROOT, &
                                 ' to ', DEST, ', tag: ', TAG
                              CALL MPITEST_MESSAGE(MPITEST_INFO1, &
                                 INFOBUF)

                              CALL MPI_SEND(BUFFER, 1, NEWTYPE, DEST, &
                                 TAG, COMM, IERR)
                              IF (IERR .NE. MPI_SUCCESS) THEN
                                 WRITE(INFOBUF, 99) &
                                    'MPI_SEND() returned ', IERR
                                 CALL MPITEST_MESSAGE( &
                                    MPITEST_NONFATAL, INFOBUF)
                                 ERRORSTRING=' '
                                 CALL MPI_ERROR_STRING(IERR, &
                                    ERRORSTRING, SIZE, IERR2)
                                 WRITE(INFOBUF, 100) ERRORSTRING(1:SIZE)
                                 CALL MPITEST_MESSAGE(MPITEST_FATAL, &
                                    INFOBUF)
                              END IF
                           ELSE IF (MPITEST_CURRENT_RANK .EQ. DEST) &
                               THEN
!                             Receive data from root's rank
!
                              WRITE(INFOBUF, 140) &
                                 'Receiving from source: ', ROOT, &
                                 ' to ', DEST, ', tag: ', TAG
                              CALL MPITEST_MESSAGE(MPITEST_INFO1, &
                                 INFOBUF)

                              CALL MPI_RECV(BUFFER, 1, NEWTYPE, ROOT, &
                                 TAG, COMM, STATUS, IERR)
                              IF (IERR .NE. MPI_SUCCESS) THEN
                                 WRITE(INFOBUF, 99) &
                                    'MPI_RECV() returned ', IERR
                                 CALL MPITEST_MESSAGE( &
                                    MPITEST_NONFATAL, INFOBUF)
                                 ERRORSTRING=' '
                                 CALL MPI_ERROR_STRING(IERR, &
                                    ERRORSTRING, SIZE, IERR2)
                                 WRITE(INFOBUF, 100) ERRORSTRING(1:SIZE)
                                 CALL MPITEST_MESSAGE(MPITEST_FATAL, &
                                    INFOBUF)
                              END IF

#ifdef MPITEST_STATUS_CHK
!                             Check status(MPI_SOURCE)
!
                              WRITE(INFOBUF, 110) &
                                 'Verifying output status object'
                              CALL MPITEST_MESSAGE(MPITEST_INFO1, &
                                 INFOBUF)

                              IF (STATUS(MPI_SOURCE) .NE. ROOT) THEN
                                 FAIL = FAIL + 1
                                 WRITE(INFOBUF, 150) &
                                    'status object returned from ', &
                                    'MPI_RECV() has unexpected ', &
                                    'MPI_SOURCE field Expected: ', &
                                    ROOT, ', Actual: ', &
                                    STATUS(MPI_SOURCE)
                                 CALL MPITEST_MESSAGE( &
                                    MPITEST_NONFATAL, INFOBUF)
                              END IF

!                             Check STATUS(MPI_TAG)
!
                              IF (STATUS(MPI_TAG) .NE. TAG) THEN
                                 FAIL = FAIL + 1
                                 WRITE(INFOBUF, 150) &
                                    'status object returned from ', &
                                    'MPI_RECV() has unexpected ', &
                                    'MPI_SOURCE field Expected: ', &
                                    TAG, ', Actual: ', &
                                    STATUS(MPI_TAG)
                                 CALL MPITEST_MESSAGE( &
                                    MPITEST_NONFATAL, INFOBUF)
                              END IF

!                             Not checking STATUS(MPI_ERROR)
!

!                             Checking the length of the message received
!
                              IF (LENGTH .NE. 0) THEN
                                 EXPECTED_COUNT = 1
                              ELSE
                                 EXPECTED_COUNT = 0
                              END IF

                              CALL MPI_GET_COUNT(STATUS, NEWTYPE, &
                                 COUNT, IERR)
                              IF (IERR .NE. MPI_SUCCESS) THEN
                                 WRITE(INFOBUF, 99) &
                                    'MPI_GET_COUNT() returned ', &
                                    IERR
                                 CALL MPITEST_MESSAGE( &
                                    MPITEST_NONFATAL, INFOBUF)
                                 ERRORSTRING=' '
                                 CALL MPI_ERROR_STRING(IERR, &
                                    ERRORSTRING, SIZE, IERR2)
                                 WRITE(INFOBUF, 100) ERRORSTRING(1:SIZE)
                                 CALL MPITEST_MESSAGE(MPITEST_FATAL, &
                                    INFOBUF)
                              END IF

                              IF (COUNT .NE. EXPECTED_COUNT) THEN
                                 FAIL = FAIL + 1
                                 WRITE(INFOBUF, 150) &
                                    'status object returned from ', &
                                    'MPI_RECV() contains ', &
                                    'unexpected length  Expected:', &
                                    EXPECTED_COUNT, ', Actual: ', &
                                    COUNT
                                 CALL MPITEST_MESSAGE( &
                                    MPITEST_NONFATAL, INFOBUF)
                              END IF
#else
                              WRITE(INFOBUF, 110) &
                                 'output status object *not* verified'
                              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, &
                                 INFOBUF)
#endif

!                             Verify the received buffer
!
                              WRITE(INFOBUF, 110) &
                                 'Verifying data received'
                              CALL MPITEST_MESSAGE( &
                                 MPITEST_INFO1, INFOBUF)

!                             Expecting to receive the root rank number
!
                              CALL MPITEST_DATATEMPLATE_INIT( &
                                 VALUE, ROOT, TEST_TYPE)
                              CALL MPITEST_BUFFER_ERRORS( &
                                 TEST_TYPE, LENGTH, VALUE, &
                                 BUFFER, ERROR)

!                             Check for receive buffer overflow
!
                              CALL MPITEST_DATATEMPLATE_INIT( &
                                 VALUE, MPITEST_CURRENT_RANK, &
                                 TEST_TYPE)
                              CALL MPITEST_BUFFER_ERRORS_OV( &
                                 TEST_TYPE, LENGTH, VALUE, &
                                 BUFFER, ERROR2)

                              IF (ERROR + ERROR2 .NE. 0) THEN
                                 IF (IERR .EQ. MPI_SUCCESS) THEN
                                    FAIL = FAIL + 1
                                 END IF
                                 WRITE(INFOBUF, 160) ERROR + ERROR2, &
                                    ' errors in buffer (', &
                                    LENGTH_COUNT, ' , ', &
                                    COMM_COUNT, ') len ', LENGTH, &
                                    ' commsize ', TEST_NUMP, &
                                    ' commtype ', COMM_TYPE, &
                                    ' data_type ', TEST_TYPE, &
                                    ' root ', ROOT
                                 CALL MPITEST_MESSAGE( &
                                    MPITEST_NONFATAL, INFOBUF)
                              ELSE
                                 WRITE(INFOBUF, 110) &
                                    '0 error found in buffer'
                                    CALL MPITEST_MESSAGE( &
                                       MPITEST_INFO2, &
                                       INFOBUF)
                              END IF
                           END IF
 600                    CONTINUE

!                       Free user defined type
!
                        CALL MPI_TYPE_FREE(NEWTYPE, IERR)
                        IF (IERR .NE. MPI_SUCCESS) THEN
                           WRITE(INFOBUF,99) &
                              'MPI_TYPE_FREE() returned ', IERR 
                           CALL MPITEST_MESSAGE(MPITEST_NONFATAL, &
                              INFOBUF)
                           ERRORSTRING=' '
                           CALL MPI_ERROR_STRING(IERR, &
                              ERRORSTRING, SIZE, IERR2)
                           WRITE(INFOBUF, 100) ERRORSTRING(1:SIZE)
                           CALL MPITEST_MESSAGE(MPITEST_FATAL, &
                              INFOBUF)
                        END IF
                     END IF
 500              CONTINUE
 400           CONTINUE
            END IF
         END IF
1000        CONTINUE

         CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM)
 300  CONTINUE

!     Report overall result
!
      CALL MPITEST_REPORT(LOOP_CNT - FAIL, FAIL, 0, TESTNAME)

      CALL MPI_FINALIZE(IERR)

      END
