! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
! **********************************************************************
!                   Test for MPI_Initialized()
! 
! This test calls MPI_Initialized() before a call to MPI_Init(), and after, and
! verifies the proper return value.
! 
! Test history:
!    1  05/02/96     gt   Created
!
! **********************************************************************
#include "f90options.h"


      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


!                           counts total number of PASSes/FAILures/VERIFY
      INTEGER LOOP_CNT
      INTEGER FAIL              
!                           return value from MPI calls
      INTEGER SAVEDIERR
      INTEGER IERR
      INTEGER ERR
      LOGICAL FLAG
!                           error class of IERR
      INTEGER ERRORCLASS
      INTEGER ERRSIZE 

      CHARACTER*(IOSIZE)  INFOBUF
      CHARACTER*32   TESTNAME
      CHARACTER*(MPI_MAX_ERROR_STRING) ERRORSTRING


 98   FORMAT(A,INT_FMT,A)
 99   FORMAT(A,INT_FMT)
 100  FORMAT(A200)


      CALL MPI_INITIALIZED(FLAG, SAVEDIERR)
!  Can't process errors until after we call MPITEST_INIT
!
!  Initialize the MPI environment and test environment.
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR 
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

      TESTNAME='MPI_Initialized'
      CALL MPITEST_INIT(IERR)
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF

      LOOP_CNT = 2
      FAIL = 0

      IF (SAVEDIERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF, 99) &
           'MPI_INITIALIZED() before MPI_INIT return code', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        INFOBUF = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
        FAIL = FAIL + 1
      END IF
      IF (FLAG .NEQV. .FALSE.) THEN
        WRITE(INFOBUF, 100) &
            'MPI_INITIALIZED() before MPI_INIT returned .TRUE.'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF


      CALL MPI_INITIALIZED(FLAG, SAVEDIERR)
      IF (SAVEDIERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF, 99) &
           'MPI_INITIALIZED() after MPI_INIT return code', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        INFOBUF = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
        FAIL = FAIL + 1
      END IF
      IF (FLAG .NEQV. .TRUE.) THEN
        WRITE(INFOBUF, 100) &
            'MPI_INITIALIZED() after MPI_INIT returned .FALSE.'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
!
!  Report overall results
!
      CALL MPITEST_REPORT(LOOP_CNT-FAIL, FAIL, 0, TESTNAME)

      CALL MPI_FINALIZE(IERR)

      END
