! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
!***********************************************************************
!  
!                          MPI_Intercomm_merge_err1
! 
!  This test verifies that the correct error is returned if 
!  MPI_Intercomm_merge() is called with invalid arguments.
!  
!  MPI_Intercomm_merge error tests
!  -----------------------------------
!  1)  Call with MPI_COMM_NULL ............................... [COMM]
!  2)  Call with an intercommunicator ........................ [ARG/COMM/OTHER]
!  3)  Call with different values of high within a  group .... [ARG/OTHER]
!  
!  In all cases, expect to receive appropriate error.
!  
!  MPI Calls dependencies for this test:
!    MPI_Intercomm_merge(), MPI_Init(), MPI_Finalize()
!    MPI_Error_string(), MPI_Comm_split(), MPI_Comm_dup(),
!    MPI_Intercomm_create(),
!***********************************************************************

#include "f90options.h"

      PROGRAM MAIN
        USE MPI
!
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

!
      INTEGER    MIN_NODES
      PARAMETER (MIN_NODES = 1)
!                                  Minimum nodes needed to run job
      CHARACTER  TEST_DESC*(*)
      PARAMETER( TEST_DESC = &
        'MPI_Intercomm_merge: call with LOCAL_COMM =  MPI_COMM_NULL')
!
      CHARACTER  ERR_DESC*(*)
      PARAMETER (ERR_DESC  = ' LOCAL_COMM =  MPI_COMM_NULL')
!
      INTEGER PASS
      INTEGER FAIL              
!                           counts total number of PASSes/FAILures
      INTEGER IERR
      INTEGER ERR
!                           return value from MPI calls
      INTEGER ERRORCLASS
      INTEGER SIZE 
!                           error class of IERR
      INTEGER RESULT 
!                           result from Intercomm_merge
      CHARACTER*(IOSIZE)  INFOBUF
!                           buffer for passing messages to MPITEST
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
!                           string describing error flag
      CHARACTER*32   TESTNAME
!                           name of this test
      INTEGER LOCAL_COMM
!                           local intra-communicator
      INTEGER PEER_COMM
!                           "peer" inter-communicator
      INTEGER INTERCOMM
!                           inter-communicator
      INTEGER NEWINTERCOMM
!                           the new inter-communicator
      INTEGER LOCAL_LEADER  
!                           rank of the local group leader in local_comm
      INTEGER REMOTE_LEADER
!                           rank of the remote leader in peer_comm
      INTEGER TAG
!                           "safe" tag
      INTEGER COLOR
!                           discriminator used by MPI_Comm_split 
      INTEGER KEY
!                           controls ordering in new LOCAL_COMMs
      LOGICAL HIGH
!                           Used in deriving the ordering within new communicator
!-----------------------------------------------------------------------------
  98  FORMAT(A,A)
 100  FORMAT(A200)
 101  FORMAT(A, A, INT_FMT)
 102  FORMAT(A, INT_FMT, A, INT_FMT)
 103  FORMAT(A, INT_FMT)
 104  FORMAT(A, INT_FMT, A)
!
!  Initialize the MPI environment and test environment.
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,103) 'MPI_INIT() returned', IERR 
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
      TESTNAME='MPI_Intercomm_merge_err1'
      CALL MPITEST_INIT(IERR)
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF

      PASS = 0
      FAIL = 0
! 
!  Set an errorhandler so we get control back.  
! 
      CALL MPI_ERRHANDLER_SET(MPI_COMM_WORLD, MPI_ERRORS_RETURN, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 103)  &
          ' MPI_ERRORHANDLER_SET(MPI_COMM_WORLD)  returned ', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

      IF(MPITEST_ME .EQ. 0) THEN
        WRITE(INFOBUF, 98)  &
        'Calling MPI_Intercomm_merge with ', ERR_DESC
        CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
      END IF
!
      HIGH = .FALSE.
!
!     Change some values to generate desired error
!
      INTERCOMM = MPI_COMM_NULL
!
!-----------------------  MPI_Intercomm_merge  -----------------------------
!
      CALL MPI_INTERCOMM_MERGE(INTERCOMM, HIGH, NEWINTERCOMM, IERR)
      IF (IERR .EQ. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 98) TEST_DESC, ' did not FAIL'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
      ELSE
        CALL MPI_ERROR_CLASS(IERR, ERRORCLASS, ERR)
        IF (ERRORCLASS .NE. MPI_ERR_COMM) THEN
          FAIL = FAIL + 1
          WRITE(INFOBUF,101) TEST_DESC, ' Returned',  ERRORCLASS 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          WRITE(INFOBUF,104)  &
            'Expected MPI_ERR_COMM(',   MPI_ERR_COMM, ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ELSE
          PASS = PASS + 1
          WRITE(INFOBUF,102) 'IERR = ', IERR, '  ERRORCLASS = ', &
            ERRORCLASS
          CALL MPITEST_MESSAGE(MPITEST_INFO2, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
        END IF
      END IF
      CALL MPITEST_REPORT(PASS, FAIL, 0, TESTNAME)
      CALL MPI_FINALIZE(IERR)
      END
