! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
!***********************************************************************
! 
!                    MPI_Comm_compare_null
! 
! This test verifies that the correct result is returned if MPI_Comm_compare()
! is called with MPI_COMM_NULL.
! 
! 10-31-02: The correct result is to indicate an error.  See MPI-1, 
! 2.4.1 on Opaque Objects and the absence of anything in 5.4.1 (Communicator
! Accessors) that allows a null handle.

! All other rank(s) will simply do nothing.
! 
! MPI Calls dependencies for this test:
!   MPI_Comm_compare(), MPI_Init(), MPI_Finalize()
!   MPI_Error_string(),
!   [MPI_Comm_compare()]
!***********************************************************************

#include "f90options.h"

      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


      CHARACTER  TEST_DESC*(*)
      PARAMETER( TEST_DESC ='MPI_Comm_compare: call with MPI_COMM_NULL')


      INTEGER PASS
      INTEGER FAIL              
!                           counts total number of PASSes/FAILures
      INTEGER IERR
      INTEGER ERR
!                           return value from MPI calls
      INTEGER ERRORCLASS
      INTEGER SIZE 
!                           error class of IERR
      INTEGER RESULT 
!                           result from Comm_compare
      CHARACTER*(IOSIZE)  INFOBUF
!                           buffer for passing messages to MPITEST
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
!                           string describing error flag
      CHARACTER*32   TESTNAME
!                           name of this test
      INTEGER   COMM1
      INTEGER   COMM2
!                           communicators
!-----------------------------------------------------------------------------
 98   FORMAT(A,A)
 100  FORMAT(A200)
 101  FORMAT(A, A, INT_FMT)
 102  FORMAT(A, INT_FMT, A, INT_FMT)
 103  FORMAT(A, INT_FMT)
 104  FORMAT(A, INT_FMT, A)
!
!  Initialize the MPI environment and test environment.
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,103) 'MPI_INIT() returned', IERR 
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

      TESTNAME='MPI_Comm_compare_null'
      CALL MPITEST_INIT(IERR)
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF

      PASS = 0
      FAIL = 0

! 
!  Set an errorhandler so we get control back.  
! 
      CALL MPI_ERRHANDLER_SET(MPI_COMM_WORLD, MPI_ERRORS_RETURN, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 103) 'MPI_ERRORHANDLER_SET returned ', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
! 
!  Set an errorhandler so we get control back.  
! 
      CALL MPI_ERRHANDLER_SET(MPI_COMM_SELF, MPI_ERRORS_RETURN, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 103) 'MPI_ERRORHANDLER_SET returned ', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

!     Set variables for call
      COMM1 = MPI_COMM_NULL
      COMM2 = MPI_COMM_WORLD

      CALL MPI_COMM_COMPARE(COMM1, COMM2, RESULT, IERR)
      IF (IERR .EQ. MPI_SUCCESS) THEN
          WRITE(INFOBUF,104) &
           'Zero return code from MPI_COMM_COMPARE 1(', &
           IERR, ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
      ELSE
        PASS = PASS + 1
      END IF

!     Set variables for call
      COMM1 = MPI_COMM_WORLD
      COMM2 = MPI_COMM_NULL

      CALL MPI_COMM_COMPARE(COMM1, COMM2, RESULT, IERR)
      IF (IERR .EQ. MPI_SUCCESS) THEN
          WRITE(INFOBUF,104) &
           'Zero return code from MPI_COMM_COMPARE 2(', &
           IERR, ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
      ELSE
        PASS = PASS + 1
      END IF

!     Set variables for call
      COMM1 = MPI_COMM_NULL
      COMM2 = MPI_COMM_SELF

      CALL MPI_COMM_COMPARE(COMM1, COMM2, RESULT, IERR)
      IF (IERR .EQ. MPI_SUCCESS) THEN
          WRITE(INFOBUF,104) &
           'Zero return code from MPI_COMM_COMPARE 3(', &
           IERR, ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
      ELSE
        PASS = PASS + 1
      END IF

!     Set variables for call
      COMM1 = MPI_COMM_SELF
      COMM2 = MPI_COMM_NULL

      CALL MPI_COMM_COMPARE(COMM1, COMM2, RESULT, IERR)
      IF (IERR .EQ. MPI_SUCCESS) THEN
          WRITE(INFOBUF,104) &
           'Zero return code from MPI_COMM_COMPARE 4(', &
           IERR, ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
      ELSE
        PASS = PASS + 1
      END IF


!     Set variables for call
      COMM1 = MPI_COMM_NULL
      COMM2 = MPI_COMM_NULL

      CALL MPI_COMM_COMPARE(COMM1, COMM2, RESULT, IERR)
      IF (IERR .EQ. MPI_SUCCESS) THEN
          WRITE(INFOBUF,104) &
           'Zero return code from MPI_COMM_COMPARE 5(', &
           IERR, ')'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
      ELSE
        PASS = PASS + 1
      END IF



      CALL MPITEST_REPORT(PASS, FAIL, 0, TESTNAME)
      CALL MPI_FINALIZE(IERR)
      END
