! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
! **********************************************************************
!  			  Test for MPI_Intercomm_create
!  
!  This test verifies that MPI_Intercomm_create combines two eperate 
!  commuicators, and that message traffic is sorted by communicator. 
!  Each configured intra- communicator (with more than 2 ranks) is split, 
!  an intercommunicator is created and tested.
!  
!  There are 2 create tests.  The difference is communicators are split the
!  following way:
!   1)  split in half
!   2)  split in thirds: rank 0 in 1 comm, ranks 2 through (size - 1) in second
!  	comm, rank 1 not in either.
!  
!  As a consequence of this test, MPI_Comm_free, MPI_Comm_compare, 
!  MPI_Comm_size, MPI_Comm_remote_size, MPI_Comm_rank, MPI_Comm_test_inter 
!  and rank are also tested here.  In addition, MPI_Comm_dup is tested with
!  intercommunicators here, also verifying that cached information in the 
!  parent communicator is inhereted by the child communicator.
!  
!  Test history:
!     1  08/28/96     gt       Original version
!  
! **********************************************************************

#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'
!
      PROGRAM MAIN
        USE MPI
!
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

!
!  Program constants (Parameters)
!
!
      CHARACTER*32          TESTNAME
      PARAMETER( TESTNAME = 'MPI_INTERCOMM_CREATE1')
!
!  General variables that occur in all programs
!
      INTEGER    CNUMBR
!                               index for communications loop
      INTEGER    COMM_INDEX
!                               the array index of the current comm
      INTEGER    COMM_TYPE
!                               the index of the current communicator type
      INTEGER    COMM_COUNT
!                               loop counter for communicator loop
      INTEGER    FTEST_NUMP
!                               number of ranks in the current communicator
      INTEGER    LOOP_COUNT
!                               counts total number of failures, loops
      INTEGER    COMM
!                               MPITEST  communicator under test
      INTEGER    IERR
      INTEGER    ERR
!                               return value from MPI calls
      INTEGER    ERRORCLASS
      INTEGER    SIZE 
!                               error class of IERR
      INTEGER    ERRSIZE
!                               length of error message
      INTEGER    PASS
      INTEGER    FAIL              
      INTEGER    VERIFY
!                           counts total number of PASSes/FAILures
      CHARACTER*(IOSIZE)    INFOBUF
!                           buffer for passing messages to MPITEST
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
!                           string describing error flag
!
!  Variables specific to this program
!
      LOGICAL    INTERCOMM
!                               intercommunicator flag
      LOGICAL    FOUND
!                               FLAG from MPI_Attr_get
      INTEGER    RESULT
!                               output from MPI_Comm_compare
      INTEGER    GROUP
!                               group handle
      INTEGER    COMMINTER
      INTEGER    COMM2
!                               for split communicator
      INTEGER    KEY
!                               MPI_Key
      INTEGER    COMMSIZE
!                               communicator size
      INTEGER    EXTRA
!                               for keyval creation/use
      INTEGER    VALUE
      INTEGER    ATTR
      INTEGER    COLOR
      INTEGER    COMMSPLIT
      INTEGER    RESULTSPLIT
      INTEGER    RESULTINTER
      INTEGER    SEND1
      INTEGER    SEND2
      INTEGER    SEND3
      INTEGER    STAT1(MPI_STATUS_SIZE)
      INTEGER    STAT2(MPI_STATUS_SIZE)
      INTEGER    STAT3(MPI_STATUS_SIZE)
      INTEGER    REQ1
      INTEGER    REQ2
      INTEGER    REQ3
      INTEGER    RECV1
      INTEGER    RECV2
      INTEGER    RECV3

!-----------------------------------------------------------------------------
 98   FORMAT(A,A )
 99   FORMAT(A,INT_FMT)
 100  FORMAT(A200)
 101  FORMAT(A, A, INT_FMT, A, INT_FMT)
 102  FORMAT(A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT , &
             A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT)
 103  FORMAT(A, INT_FMT, A, A, A, INT_FMT)
 104  FORMAT(A, A, A, INT_FMT, A)
 105  FORMAT(A, INT_FMT, A, A, INT_FMT, A)
!===========================================================================
!
!  Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!  Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
!
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF
!
!  Get number of communicators
!
      CALL MPITEST_NUM_COMM_SIZES(COMM_COUNT)

      FAIL = 0
      LOOP_COUNT = 0
      VERIFY = 0
!
!-------------------------  Create a Keyval  ----------------------------
!

      CALL   MPI_KEYVAL_CREATE(MPI_DUP_FN, &
                               MPI_NULL_DELETE_FN, &
                               KEY, &
                               EXTRA, &
                               IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF,103) 'Non-Zero return code (', IERR,  &
          ')  From:  MPI_KEYVAL_CREATE'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
      END IF
!
!---------------------  Loop over Communicators  ------------------------
!
      DO 4000 CNUMBR = 1,COMM_COUNT
        CALL MPITEST_GET_COMM_INDEX(CNUMBR, COMM_INDEX)
        CALL MPITEST_GET_COMM_TYPE(CNUMBR, COMM_TYPE)
        CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, COMM, &
              FTEST_NUMP)
!
!       Skip to end of loop if this node is not part of current
!       communicator
!
        IF (MPITEST_CURRENT_RANK .EQ. MPI_UNDEFINED ) GO TO 3800

!
!     Check for intercommunicator and skip (continue) if so
!     An inter-communicator cannot be split  
!
        CALL MPI_COMM_TEST_INTER(COMM,  &
                                 INTERCOMM,  &
                                 IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_TEST_INTER  ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1

        IF((INTERCOMM .EQV. .TRUE.)  .OR.  (FTEST_NUMP .LT. 2)) THEN
          CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM)
          GO TO 4000
        END IF
!
!     Split the communicator in half
!
        COLOR = MOD(MPITEST_CURRENT_RANK,2)
!
        CALL MPI_COMM_SPLIT(COMM, &
                            COLOR, &
                            MPITEST_ME, &
                            COMMSPLIT,  &
                            IERR)
!
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_SPLIT ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Put some messages on the communicator before creating
!  the inter-comm to ensure the correct TAG is used
!
        IF(MPITEST_CURRENT_RANK .LT. 2) THEN
          SEND1 = 19
          SEND2 = 29
          SEND3 = 29
          CALL MPI_ISEND(SEND1,  &
                         1,  &
                         MPI_INTEGER,  &
                         1 - MPITEST_CURRENT_RANK,  &
                         26,  &
                         COMM,   &
                         REQ1,  &
                         IERR)
!
          CALL MPI_ISEND(SEND2,  &
                         1,  &
                         MPI_INTEGER,  &
                         1 - MPITEST_CURRENT_RANK,  &
                         28,  &
                         COMM,   &
                         REQ2,  &
                         IERR)
!
          CALL MPI_ISEND(SEND3,  &
                         1,  &
                         MPI_INTEGER,  &
                         1 - MPITEST_CURRENT_RANK,  &
                         0,  &
                         COMM,   &
                         REQ3,  &
                         IERR)
        END IF
!
!     Now combine the two new communicators to create an intercommunicator
!
        CALL MPI_INTERCOMM_CREATE(COMMSPLIT, &
                                  0, &
                                  COMM, &
                                  1 - COLOR,  &
                                  27,  &
                                  COMMINTER,  &
                                  IERR)
!
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_INTERCOMM_CREATE ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Now receive the messages from above, and ensure they are intact
!
        IF(MPITEST_CURRENT_RANK .LT. 2) THEN
          CALL MPI_RECV(RECV1,  &
                        1,  &
                        MPI_INTEGER,  &
                        1 - MPITEST_CURRENT_RANK,  &
                        26,  &
                        COMM,   &
                        STAT1,  &
                        IERR)
!
          CALL MPI_RECV(RECV2,  &
                        1,  &
                        MPI_INTEGER,  &
                        1 - MPITEST_CURRENT_RANK,  &
                        28,  &
                        COMM,   &
                        STAT2,  &
                        IERR)
!
          CALL MPI_RECV(RECV3,  &
                        1,  &
                        MPI_INTEGER,  &
                        1 - MPITEST_CURRENT_RANK,  &
                        0,  &
                        COMM,   &
                        STAT3,  &
                        IERR)
!
          CALL MPI_WAIT(REQ1, STAT1, IERR)
          CALL MPI_WAIT(REQ2, STAT2, IERR)
          CALL MPI_WAIT(REQ3, STAT3, IERR)
!
          IF(RECV1 .NE. SEND1  .OR.  RECV2 .NE. SEND2  .OR.   &
             RECV3 .NE. SEND3) THEN

            WRITE(INFOBUF,102)  &
            'Receive data corrupted by MPI_Intercomm_create:  ', &
              RECV1, '/', RECV2, '/', RECV3,   &
              '  Expected:  ', SEND1, '/', SEND2, '/', SEND3,   &
            ' Comm_Index = ', COMM_INDEX 
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            FAIL = FAIL + 1
          END IF
          LOOP_COUNT = LOOP_COUNT + 1
        END IF

!
!  Compare the two communicators
!
        CALL MPI_COMM_COMPARE(COMM,  &
                              COMMINTER,  &
                              RESULT,  &
                              IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR, &
            ')  From:  MPI_COMM_COMPARE ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF

          IF(RESULT .NE. MPI_UNEQUAL) THEN
            WRITE(INFOBUF,105)  &
              'MPI_COMM_COMPARE returned', RESULT,  &
              ',  expected  MPI_UNEQUAL',  &
              ' Comm_Index = ', COMM_INDEX 
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Set a value for the attribute
!
        VALUE = 27
        LOOP_COUNT = LOOP_COUNT + 1
!
        CALL MPI_ATTR_PUT(COMMINTER,  &
                          KEY,  &
                          VALUE,  &
                          IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_ATTR_PUT ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        END IF
!
!  Get the value for the attribute; it should be there
!
            CALL MPI_ATTR_GET( COMMINTER, &
                               KEY, &
                               ATTR, &
                               FOUND, &
                               IERR)
            IF (IERR .NE. MPI_SUCCESS) THEN
              WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
                ')  From:  MPI_Attr_get(COMMINTER) '
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              ERRORSTRING = ' '
              CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
              WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              FAIL = FAIL + 1
            ELSE IF  (FOUND .NEQV. .TRUE.) THEN
              WRITE(INFOBUF,104)  &
              'MPI_Attr_get(COMMINTER) did not find attribute ',  &
              ' (', 'COMM_INDEX', COMM_INDEX, ')' 
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              FAIL = FAIL + 1
            ELSE IF  (ATTR .NE. 27) THEN
              WRITE(INFOBUF,105)  &
                'MPI_Attr_get(COMMINTER) returned value ', ATTR, &
                '  Expected  27 (', &
                'COMM_INDEX', COMM_INDEX, ')' 
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              FAIL = FAIL + 1
            END IF
            LOOP_COUNT = LOOP_COUNT + 1

!
!     Check for inter-communicator -- We know it is
!
        CALL MPI_COMM_TEST_INTER(COMMINTER,  &
                                 INTERCOMM,  &
                                 IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_TEST_INTER #2 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1

        IF(INTERCOMM .NEQV. .TRUE.) THEN
          WRITE(INFOBUF,102)  &
            'MPI_COMM_TEST_INTER  did NOT return TRUE ( ', INTERCOMM, &
            ')   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF

!
!  Check the Size
!
        CALL MPI_COMM_SIZE(COMMINTER,  &
                           RESULTINTER,  &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_SIZE ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1

!  Size should be half the original, plus one for odd rank total, 
!  and this is an even-numbered rank
!

        RESULT = FTEST_NUMP/2
        IF(MOD(FTEST_NUMP,2) .EQ. 1  .AND.   &
             MOD(MPITEST_CURRENT_RANK, 2) .EQ. 0) RESULT = RESULT + 1
!
        IF ( RESULTINTER .NE. RESULT) THEN
          WRITE(INFOBUF,102)  &
            'MPI_COMM_SIZE returned', RESULTINTER,  &
            ',  Expected ',  RESULT, &
            '   Comm_Index = ', COMM_INDEX 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!
!  Check the REMOTE Size
!
        CALL MPI_COMM_REMOTE_SIZE(COMMINTER,  &
                                  RESULTINTER,  &
                                  IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_REMOTE_SIZE ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        RESULT = FTEST_NUMP - RESULT
!
        IF ( RESULTINTER .NE. RESULT) THEN
          WRITE(INFOBUF,102)  &
            'MPI_COMM_REMOTE_SIZE returned', RESULTINTER,  &
            ',  Expected ',  RESULT, &
            '   Comm_Index = ', COMM_INDEX 
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!
!  Check the rank, it should be the same value as in the split communicator 
!
        CALL MPI_COMM_RANK(COMMINTER, &
                           RESULTINTER,  &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Comm_rank ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1

        IF ( RESULTINTER .NE. MPITEST_CURRENT_RANK/2) THEN
          WRITE(INFOBUF,102)  &
            'MPI_COMM_RANK returned', RESULTINTER,  &
            ',  Expected ',  MPITEST_CURRENT_RANK/2, &
            '   Comm_Index = ', COMM_INDEX 
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
!
! Resultsplit is nowhere set!
!C        ELSE IF (RESULTSPLIT .NE. 0) THEN
!C          WRITE(INFOBUF,105) 'MPI_Comm_rank returned', RESULTSPLIT, 
!C     $      ',  expected  0 ', 
!C     $      ' Comm_Index = ', COMM_INDEX 
!C          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
!C          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Try the new communicator
!
!     This appears to be old code from the MPI_Comm_split case that was
!     never removed (but was used to perform the split used to 
!     create the intercommunicator)
!C
!C        IF(RESULTSPLIT .EQ. 0) RESULT = COLOR
!CC
!C        CALL MPI_BCAST( RESULT, 
!C     $                    1, 
!C     $                    MPI_INTEGER, 
!C     $                    0, 
!C     $                    COMMSPLIT,
!C     $                    IERR )
!C        IF (IERR .NE. MPI_SUCCESS) THEN
!C          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR, 
!C     $      ')  From:  MPI_Bcast '
!C          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
!C          ERRORSTRING = ' '
!C          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
!C          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
!C          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
!C          FAIL = FAIL + 1
!C        ELSE IF  (RESULT .NE. COLOR) THEN
!C          FAIL = FAIL + 1
!C          WRITE(INFOBUF,102) 
!C     $      'MPI_Bcast received incorrect data, Actual = ', RESULT,
!C     $      '  Expected', COLOR, 
!C     $      '  COMM_INDEX = ', COMM_INDEX 
!C          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
!C        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!     
!     Try to dup the new intercommunicator
!
        CALL MPI_COMM_DUP(COMMINTER,  &
                          COMM2,  &
                          IERR)
!
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_DUP ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Get the value for the attribute; it should be there
!
        CALL MPI_ATTR_GET( COMM2, &
                               KEY, &
                               ATTR, &
                               FOUND, &
                               IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(COMM2) '
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV. .TRUE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(COMM2) did not find attribute ',  &
            ' (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (ATTR .NE. 27) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(COMM2) returned value ', ATTR, &
            '  Expected  27', &
            ' (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Delete the Attribute
!
        CALL MPI_ATTR_DELETE(COMM2, &
                                 KEY,  &
                                 IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_ATTR_DELETE after delete ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!     
!     Compare the two INTER-communicators
!
        CALL MPI_COMM_COMPARE(COMMINTER,  &
                                  COMM2,  &
                                  RESULT,  &
                                  IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR, &
            ')  From:  MPI_COMM_COMPARE #2 ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
        
        IF(RESULT .NE. MPI_CONGRUENT) THEN
          WRITE(INFOBUF,102)  &
            'MPI_COMM_COMPARE #2 for one rank  returned', RESULT,  &
            ',  expected  MPI_CONGRUENT(', MPI_CONGRUENT, &
            ')   Comm_Index = ', COMM_INDEX 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Free all  Communicators
!
        CALL MPI_COMM_FREE(COMMSPLIT, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_FREE(COMMSPLIT)  ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        IF(COMMSPLIT .NE. MPI_COMM_NULL)  THEN
          WRITE(INFOBUF,105)  &
            'COMMSPLIT = ', COMMSPLIT,  ' after MPI_Comm_free',  &
            ',  expected MPI_COMM_NULL ( Comm_Index = ', &
            COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!     
        CALL MPI_COMM_FREE(COMMINTER, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_FREE(COMMINTER)  ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!     
        IF(COMMINTER .NE. MPI_COMM_NULL)  THEN
          WRITE(INFOBUF,105)  &
            'COMMINTER = ', COMMINTER,  ' after MPI_Comm_free',  &
            ',  expected MPI_COMM_NULL ( Comm_Index = ', &
            COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        CALL MPI_COMM_FREE(COMM2, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_FREE(COMM2)  ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        IF(COMM2 .NE. MPI_COMM_NULL)  THEN
          WRITE(INFOBUF,105)  &
            'COMM2 = ', COMM2,  ' after MPI_Comm_free',  &
            ',  expected MPI_COMM_NULL ( Comm_Index = ', &
            COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM)
 3800   CONTINUE
!                Node defined in current communicator
 4000 CONTINUE
!                End communicator loop
!
!  Free the Keyval
!
      CALL   MPI_KEYVAL_FREE(KEY, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
          ')  From:  MPI_Keyval_free'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
      END IF
!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_COUNT - FAIL, FAIL, 0, TESTNAME)
      
      CALL MPI_FINALIZE(IERR)

      END
