! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
! **********************************************************************
!  			  Test for MPI_Intercomm_merge
!  
!  This test verifies that MPI_Intercomm_create combines two eperate 
!  commuicators,  and that message traffic is sorted by communicator.  
!  Each configured intra-communicator (with more than 2 ranks) is split, 
!  an intercommunicator is created, merged, and tested.
!  
!  There are 3 merge tests.  The difference is communicators are split the
!  following way:
!   1)  split in half
!   2)  split in thirds: rank 0 in 1 comm, ranks 2 through (size - 1) in second
!          comm, rank 1 not in either. (rank 0 is "high").
!   3)  same as 2, except for order of merge (rank 0 is "low").
!  
!  As a consequence of this test, MPI_Comm_free, MPI_Comm_compare, 
!  MPI_Comm_size, MPI_Comm_rank, MPI_Comm_test_inter are also tested here.
!  
!  Test history:
!     1  08/28/96     gt       Original version
!  
! **********************************************************************

#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'
!
      PROGRAM MAIN
        USE MPI
!
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

!
!  Program constants (Parameters)
!
!
      CHARACTER*32          TESTNAME
      PARAMETER( TESTNAME = 'MPI_INTERCOMM_MERGE2')
!
!  General variables that occur in all programs
!
      INTEGER    CNUMBR
!                               index for communications loop
      INTEGER    COMM_INDEX
!                               the array index of the current comm
      INTEGER    COMM_TYPE
!                               the index of the current communicator type
      INTEGER    COMM_COUNT
!                               loop counter for communicator loop
      INTEGER    FTEST_NUMP
!                               number of ranks in the current communicator
      INTEGER    LOOP_COUNT
!                               counts total number of failures, loops
      INTEGER    COMM
!                               MPITEST  communicator under test
      INTEGER    IERR
      INTEGER    ERR
!                               return value from MPI calls
      INTEGER    ERRORCLASS
      INTEGER    SIZE 
!                               error class of IERR
      INTEGER    ERRSIZE
!                               length of error message
      INTEGER    PASS
      INTEGER    FAIL              
      INTEGER    VERIFY              
!                           counts total number of PASSes/FAILures
      CHARACTER*(IOSIZE)    INFOBUF
!                           buffer for passing messages to MPITEST
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
!                           string describing error flag
!
!  Variables specific to this program
!
      LOGICAL    INTERCOMM
!                               intercommunicator flag
      LOGICAL    FOUND
!                               FLAG from MPI_Attr_get
      INTEGER    RESULT
!                               output from MPI_Comm_compare
      INTEGER    GROUP
!                               group handle
      INTEGER    COMMINTER
      INTEGER    COMM2
!                               for split communicator
      INTEGER    KEY
!                               MPI_Key
      INTEGER    COMMSIZE
!                               communicator size
      INTEGER    EXTRA
!                               for keyval creation/use
      INTEGER    COLOR
      INTEGER    COMMSPLIT
      INTEGER    RESULTSPLIT
      INTEGER    RESULTMERGE

      INTEGER    BUFFER1
      INTEGER    BUFFER2
      INTEGER    BUFFER3
!
!-----------------------------------------------------------------------------
 98   FORMAT(A,A )
 99   FORMAT(A,INT_FMT)
 100  FORMAT(A200)
 101  FORMAT(A, A, INT_FMT, A, INT_FMT)
 102  FORMAT(A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT , &
             A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT)
 103  FORMAT(A, INT_FMT, A, A, A, INT_FMT)
 104  FORMAT(A, A, A, INT_FMT, A)
 105  FORMAT(A, INT_FMT, A, A, INT_FMT, A)
!===========================================================================
!
!  Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!  Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
!
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF
!
!  Get number of communicators
!
      CALL MPITEST_NUM_COMM_SIZES(COMM_COUNT)
!
      FAIL = 0
      LOOP_COUNT = 0
      VERIFY = 0
!
!---------------------  Loop over Communicators  ------------------------
!
      DO 4000 CNUMBR = 1,COMM_COUNT
        CALL MPITEST_GET_COMM_INDEX(CNUMBR, COMM_INDEX)
        CALL MPITEST_GET_COMM_TYPE(CNUMBR, COMM_TYPE)
        CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, COMM, &
              FTEST_NUMP)
!
!       Skip to end of loop if this node is not part of current
!       communicator
!
        IF (MPITEST_CURRENT_RANK .EQ. MPI_UNDEFINED ) GO TO 3800
!
!     Check for intercommunicator and skip (continue) if so
!     An inter-communicator cannot be split  
!
        CALL MPI_COMM_TEST_INTER(COMM,  &
                                 INTERCOMM,  &
                                 IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_TEST_INTER  ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        IF((INTERCOMM .EQV. .TRUE.)  .OR.  (FTEST_NUMP .LT. 3)) THEN
          CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM)
          GO TO 4000
        END IF
!
!     Split the communicator 
!
        COLOR = 0
        IF(MPITEST_CURRENT_RANK .EQ. 0) THEN
          COLOR = FTEST_NUMP - 1
        ELSE IF(MPITEST_CURRENT_RANK .EQ. 1) THEN
          COLOR = MPI_UNDEFINED
        END IF
!
        CALL MPI_COMM_SPLIT(COMM, &
                            COLOR, &
                            MPITEST_ME, &
                            COMMSPLIT,  &
                            IERR)
!
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_SPLIT ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
        IF(MPITEST_CURRENT_RANK .EQ. 1) THEN
          IF(COMMSPLIT .NE. MPI_COMM_NULL) THEN
            WRITE(INFOBUF,105)  &
            'MPI_Comm_split w/color = MPI_UNDEFINED returned ', &
            COMMSPLIT,   &
            '  Expected:  MPI_COMM_NULL',   &
            ' Comm_Index = ', COMM_INDEX 
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            FAIL = FAIL + 1
          END IF
          GO TO 3600
        END IF
!
!  Combine the two new communicators to create an intercommunicator
!
!     First, split 1 rank
!     0 Must be local leader, last rank is remote leader 

        IF(MPITEST_CURRENT_RANK .EQ. 0) THEN
          CALL MPI_INTERCOMM_CREATE(COMMSPLIT, &
                                    0, &
                                    COMM, &
                                    FTEST_NUMP - 1,  &
                                    65534,  &
                                    COMMINTER,  &
                                    IERR)
        ELSE
!
!     Second split, all ranks but 2
!     last rank is local leader, 0 is remote leader 

          CALL MPI_INTERCOMM_CREATE(COMMSPLIT, &
                                    FTEST_NUMP - 3,  &
                                    COMM, &
                                    0, &
                                    65534,  &
                                    COMMINTER,  &
                                    IERR)
!
        END IF

        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_INTERCOMM_CREATE ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Merge them back together again 
!
!                                             HIGH = .TRUE.
        IF(MPITEST_CURRENT_RANK .EQ. 0) THEN
          CALL MPI_INTERCOMM_MERGE(COMMINTER, &
                                   .TRUE., &
                                   COMM2, &
                                   IERR)
        ELSE
!                                             HIGH = .FALSE.
          CALL MPI_INTERCOMM_MERGE(COMMINTER, &
                                   .FALSE., &
                                   COMM2, &
                                   IERR)
!
      END IF
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
          ')  From:  MPI_INTERCOMM_MERGE ( COMM_INDEX ',  &
          COMM_INDEX
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
      LOOP_COUNT = LOOP_COUNT + 1
!
!  Compare the merged communicator with the original inter-communicator
!
        CALL MPI_COMM_COMPARE(COMM2,  &
                              COMMINTER,  &
                              RESULTMERGE,  &
                              IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR, &
            ')  From:  MPI_COMM_COMPARE ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF

        IF(RESULTMERGE .NE. MPI_UNEQUAL)  THEN 
          WRITE(INFOBUF,105)  &
            'MPI_COMM_COMPARE returned', RESULT,  &
            ',  expected  MPI_UNEQUAL',  &
            ' Comm_Index = ', COMM_INDEX 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Check for inter-communicator -- We know it isn't
!
        CALL MPI_COMM_TEST_INTER(COMM2,  &
                                 INTERCOMM,  &
                                 IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_TEST_INTER/MERGED ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1

        IF(INTERCOMM .NEQV. .FALSE.) THEN
          WRITE(INFOBUF,102)  &
            'MPI_COMM_TEST_INTER  did NOT return FALSE ( ', INTERCOMM, &
            ')   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!
!  Check the Size
!
        CALL MPI_COMM_SIZE(COMM2,  &
                           RESULTMERGE,  &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_SIZE ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Size should FTEST_NUMP - 1 (all but rank 1)
!
        IF ( RESULTMERGE .NE. FTEST_NUMP - 1) THEN
          WRITE(INFOBUF,102)  &
            'MPI_COMM_SIZE returned', RESULTMERGE,  &
            ',  Expected ',  FTEST_NUMP-1, &
            '   Comm_Index = ', COMM_INDEX 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF

!  Check the rank, it should be 0 thru FTEST_NUMP - 1
!  Remember, we put rank 0 high = FALSE
!
        CALL MPI_COMM_RANK(COMM2, &
                           RESULTMERGE,  &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Comm_rank ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1

        RESULT = FTEST_NUMP - 2
        IF(MPITEST_CURRENT_RANK .GT. 0)  &
                         RESULT = MPITEST_CURRENT_RANK - 2

        IF ( RESULTMERGE  .NE. RESULT) THEN
          WRITE(INFOBUF,102)  &
            'MPI_COMM_RANK returned ', RESULTMERGE, &
            '  Expected = ', RESULT, &
            '   Comm_Index = ', COMM_INDEX 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!
!  Try the new communicator
!
        IF(RESULTMERGE .EQ. 1) RESULT = -35
!
          CALL MPI_BCAST( RESULT,  &
                          1,  &
                          MPI_INTEGER,  &
                          1,  &
                          COMM2, &
                          IERR )
          IF (IERR .NE. MPI_SUCCESS) THEN
            WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
              ')  From:  MPI_Bcast '
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            ERRORSTRING = ' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            FAIL = FAIL + 1
          ELSE IF  (RESULT .NE. -35) THEN
            FAIL = FAIL + 1
            WRITE(INFOBUF,105)  &
              'MPI_Bcast received incorrect data, Actual = ', RESULT, &
              '  Expected 127',  &
              '  COMM_INDEX = ', COMM_INDEX 
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          END IF
          LOOP_COUNT = LOOP_COUNT + 1
!
!  Free all  Communicators
!
          CALL MPI_COMM_FREE(COMMINTER, IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
              ')  From:  MPI_COMM_FREE(COMMINTER)  ( COMM_INDEX ',  &
              COMM_INDEX
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            ERRORSTRING = ' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            FAIL = FAIL + 1
          END IF
          LOOP_COUNT = LOOP_COUNT + 1
!
          IF(COMMINTER .NE. MPI_COMM_NULL)  THEN
            WRITE(INFOBUF,105)  &
              'COMMINTER = ', COMMINTER,  ' after MPI_Comm_free',  &
              ',  expected MPI_COMM_NULL ( Comm_Index = ', &
              COMM_INDEX, ')' 
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            FAIL = FAIL + 1
          END IF
          LOOP_COUNT = LOOP_COUNT + 1
!
          CALL MPI_COMM_FREE(COMM2, IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
              ')  From:  MPI_COMM_FREE(COMM2)  ( COMM_INDEX ',  &
              COMM_INDEX
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            ERRORSTRING = ' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            FAIL = FAIL + 1
          END IF
          LOOP_COUNT = LOOP_COUNT + 1
!
          IF(COMM2 .NE. MPI_COMM_NULL)  THEN
            WRITE(INFOBUF,105)  &
              'COMM2 = ', COMM2,  ' after MPI_Comm_free',  &
              ',  expected MPI_COMM_NULL ( Comm_Index = ', &
              COMM_INDEX, ')' 
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            FAIL = FAIL + 1
          END IF
          LOOP_COUNT = LOOP_COUNT + 1
!
 3600     CONTINUE
!     Rank not 1 in this test
!
        CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM)
 3800   CONTINUE
!                Node defined in current communicator
 4000 CONTINUE
!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_COUNT - FAIL, FAIL, 0, TESTNAME)
      
      CALL MPI_FINALIZE(IERR)

      END
