! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
!***********************************************************************
!  		  Test for MPI keyvals and attributes.
!  
!  This test verifies that errors from copy or delete functions are returned to
!  an application.  It checks that the correct error is returned.
!  
!  Test history:
!     1  08/28/96     gt       Original version
!  
! NOTE:  9/96
!    The MPI Standard(6/95), Section 5.7.1(Page 169)  gives a prototype
!    for a Copy callback function (SUBROUTINE COPY_FUNCTION), and
!    a Callback deletion function (SUBROUTINE DELETE_FUNCTION).
!    The last formal parameter in both of these subroutines is an
!    INTEGER error variable, IERR.  This variable is not in the INTEL MPI
!    code, and when run with the IERR parameter included in the declaration,
!    the program crashes.  This program references Callback-function
!    calls to subroutines at the end of this file 
!***********************************************************************

#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'
!
      PROGRAM MAIN
        USE MPI
!
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

!
!  Program constants (Parameters)
!
!
      CHARACTER*32          TESTNAME
      PARAMETER( TESTNAME = 'MPI_KEYVAL3')
!
!  General variables that occur in all programs
!
      INTEGER    LOOP_COUNT
!                               counts total number of failures, loops
      INTEGER    IERR
      INTEGER    ERR
!                               return value from MPI calls
      INTEGER    ERRORCLASS
      INTEGER    SIZE 
!                               error class of IERR
      INTEGER    ERRSIZE
!                               length of error message
      INTEGER    PASS
      INTEGER    FAIL              
      INTEGER    VERIFY
!                           counts total number of PASSes/FAILures/Verifies
      CHARACTER*(IOSIZE)    INFOBUF
!                           buffer for passing messages to MPITEST
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
!                           string describing error flag
!
!  Variables specific to this program
!
      LOGICAL    INTERCOMM
!                               intercommunicator flag
      LOGICAL    FOUND
!                               FLAG from MPI_Attr_get
      INTEGER    COMM1
      INTEGER    COMM2
!                               for split communicator
      INTEGER    VALUE
      INTEGER    ATTR
      INTEGER    KEY
!                               MPI_Key
      INTEGER    EXTRA
!                               for keyval creation/use
      EXTERNAL   COPY_FUNCTION
      EXTERNAL   DELETE_FUNCTION
!-----------------------------------------------------------------------------
 98   FORMAT(A,A )
 99   FORMAT(A,INT_FMT)
 100  FORMAT(A200)
 101  FORMAT(A, A, INT_FMT, A, INT_FMT)
 102  FORMAT(A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT , &
             A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT)
 103  FORMAT(A, INT_FMT, A, A, A, INT_FMT)
 104  FORMAT(A, A, A, INT_FMT, A)
 105  FORMAT(A, INT_FMT, A, A, INT_FMT, A)
!===========================================================================
!
!  Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!  Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
!
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF

      FAIL = 0
      LOOP_COUNT = 0
      VERIFY = 0
      EXTRA = 0
!
!-------------------------  Create a Keyval  ----------------------------
!

      CALL   MPI_KEYVAL_CREATE(COPY_FUNCTION, &
                               DELETE_FUNCTION, &
                               KEY, &
                               EXTRA, &
                               IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF,103) 'Non-Zero return code (', IERR,  &
          ')  From:  MPI_KEYVAL_CREATE'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
      END IF
!
!     Dup the COMM_WORLD, for use as test communicator
!
        CALL MPI_COMM_DUP(MPI_COMM_WORLD,  &
                          COMM1,  &
                          IERR)
!
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_DUP '
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!     
!     Set a value for the attribute
!
      VALUE = 0
!
      CALL MPI_ATTR_PUT(COMM1,  &
                        KEY, &
                        VALUE,  &
                        IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
          ')  From:  MPI_ATTR_PUT   ', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
      END IF
!
!     Try to DUP it, it should not be allowed
!
        CALL MPI_COMM_DUP(COMM1,  &
                          COMM2,  &
                          IERR)
!
        IF (IERR .NE. MPI_ERR_OTHER) THEN
          WRITE(INFOBUF,102) 'Unexpected  return code (', IERR,  &
            ')  From:  MPI_COMM_DUP -- Expected MPI_ERR_OTHER'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Delete the Attribute, should not be allowed
!
      CALL MPI_ATTR_DELETE(COMM1, &
                           KEY,  &
                           IERR)
!
      IF (IERR .NE. MPI_ERR_OTHER) THEN
        WRITE(INFOBUF,102) 'Unexpected  return code (', IERR,  &
          ')  From:  MPI_ATTR_DELETE -- Expected MPI_ERR_OTHER'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
      LOOP_COUNT = LOOP_COUNT + 1
!
!  Set a NEW value for the attribute, should not be allowed
!
      VALUE = 1
!     
      CALL MPI_ATTR_PUT(COMM1,  &
                        KEY,  &
                        VALUE,  &
                        IERR)

      IF (IERR .NE. MPI_ERR_OTHER) THEN
        WRITE(INFOBUF,102) 'Unexpected  return code (', IERR,  &
          ')  From:  MPI_ATTR_PUT -- Expected MPI_ERR_OTHER'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
      LOOP_COUNT = LOOP_COUNT + 1
!
!  Try to free the COMM,  it should not be allowed
!
      CALL MPI_COMM_FREE(COMM1, IERR)
      IF (IERR .NE. MPI_ERR_OTHER) THEN
        WRITE(INFOBUF,102) 'Unexpected  return code (', IERR,  &
          ')  From:  MPI_COMM_FREE -- Expected MPI_ERR_OTHER'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
      LOOP_COUNT = LOOP_COUNT + 1
!
!  Delete the Attribute
!
      EXTRA = 2
      CALL MPI_ATTR_DELETE(COMM1, &
                           KEY,  &
                           IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
          ')  From:  MPI_ATTR_DELETE after free '

        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
!
!
!  Try to free the COMM,  it should  be allowed
!
      CALL MPI_COMM_FREE(COMM1, &
                         IERR)
!
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
          ')  From:  MPI_COMM_FREE(COMM1) '

        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
      LOOP_COUNT = LOOP_COUNT + 1
!
!  Free the Keyval
!
      CALL   MPI_KEYVAL_FREE(KEY, &
                             IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
          ')  From:  MPI_Keyval_free(2)'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
      END IF

      IF(KEY .NE. MPI_KEYVAL_INVALID) THEN
        WRITE(INFOBUF,104)  &
          'Key not set to MPI_KEYVAL_INVALID by MPI_KEYVAL_FREE'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
      LOOP_COUNT = LOOP_COUNT + 1
!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_COUNT - FAIL, FAIL, 0, TESTNAME)
      
      CALL MPI_FINALIZE(IERR)
      
      END
!------------------------------------------------------------------------------
      SUBROUTINE COPY_FUNCTION(OLDCOMM, &
                                KEYVAL, &
                                EXTRA_STATE, &
                                ATTRIBUTE_VAL_IN, &
                                ATTRIBUTE_VAL_OUT, &
                                FLAG, IERR)
!

      INTEGER  OLDCOMM
      INTEGER  KEYVAL
      INTEGER  EXTRA_STATE
      INTEGER  ATTRIBUTE_VAL_IN
      INTEGER  ATTRIBUTE_VAL_OUT
      INTEGER  IERR
!
      LOGICAL  FLAG
!
      IERR = MPI_ERR_OTHER
      RETURN
      END

!------------------------------------------------------------------------------
      SUBROUTINE DELETE_FUNCTION(COMM, &
                                 KEYVAL, &
                                 ATTRIBUTE_VAL, &
                                 EXTRA_STATE, IERR)


      INTEGER  COMM
      INTEGER  KEYVAL
      INTEGER  ATTRIBUTE_VAL
      INTEGER  EXTRA_STATE
      INTEGER  IERR
!
      IF(EXTRA_STATE .EQ. 0) THEN
        IERR = MPI_ERR_OTHER
      ELSE
	IERR = MPI_SUCCESS
      END IF
!
      RETURN
      END

