! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
!***************************************************************************
!
!                          MPI_DIMS_CREATE_ERR4
!
!
!  Reference:  MPI Standard,  Chapter 6:  Process Topologies
!                             Section 6.5.2:  Cartesian Convenience Function
!
!  MPI_Dims_create error tests
!  -----------------------------------
!  1)  Nodes not a multiple of non-zero dimensions   ..... [DIMS/ARG/OTHER]
!  2)  Negative value in dimensions array  ............... [DIMS/ARG/OTHER]
!  3)  Nodes < 0 ......................................... [ARG/OTHER]
!  4)  Number of dimensions < 0  ......................... [DIMS/ARG/OTHER]
!
!***************************************************************************

#include "f90options.h"

      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


      INTEGER    MPITEST_MAXDIMS
      INTEGER    MPITEST_DIMS
      INTEGER    MPITEST_MAXNODE
      PARAMETER( MPITEST_MAXDIMS = 10)
      PARAMETER( MPITEST_DIMS    =  4)
      PARAMETER( MPITEST_MAXNODE = 120)
      CHARACTER  TEST_DESC*(*)
      PARAMETER( TEST_DESC = 'MPI_DIMS_CREATE:  illegal node count')


!                           counts total number of PASSes/FAILures/VERIFY
      INTEGER PASS
      INTEGER FAIL              
!                           return value from MPI calls
      INTEGER IERR
      INTEGER ERR
!                           error class of IERR
      INTEGER ERRORCLASS
      INTEGER SIZE 

      INTEGER I
!                           general loop indices
      INTEGER NUM_DIMS
!                           number of dimensions to generate coordinates for
      INTEGER NNODES
!                           the number of Cartesian processes  

      CHARACTER*(IOSIZE)  INFOBUF
!                           buffer for passing messages to MPITEST
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
!                           string describing error flag
      CHARACTER*32   TESTNAME
!                           name of this test
      INTEGER COMM_CART
!                           Cartesian communicator
      INTEGER DIM_SIZE(MPITEST_MAXDIMS)
!                           processes in each dimension
!-----------------------------------------------------------------------------

 98   FORMAT(A,A)
 99   FORMAT(A,INT_FMT)
 100  FORMAT(A200)
 101  FORMAT(A, A, INT_FMT)
 102  FORMAT(A, INT_FMT, A, INT_FMT, A, INT_FMT, A , INT_FMT)
 103  FORMAT(A, INT_FMT, A, A)
!
!  Initialize the MPI environment and test environment.
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR 
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

      TESTNAME='MPI_DIMS_CREATE_ERR4'
      CALL MPITEST_INIT(IERR)
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF

      PASS = 0
      FAIL = 0

! 
!  Set an errorhandler so we get control back.  
! 
      CALL MPI_ERRHANDLER_SET(MPI_COMM_WORLD, MPI_ERRORS_RETURN, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 99) ' MPI_ERRORHANDLER_SET returned ', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

      NUM_DIMS = MPITEST_DIMS
!
!--------------------------  MPI_CART_CREATE  ------------------------
!
      NNODES = 120
      NUM_DIMS = -1

      DIM_SIZE(1) = 4
      DIM_SIZE(2) = 5
      DIM_SIZE(3) = 1
      DIM_SIZE(4) = 6

      CALL MPI_DIMS_CREATE( NNODES, &
                            NUM_DIMS, &
                            DIM_SIZE, &
                            IERR)

      IF (IERR .EQ. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 98) TEST_DESC, ' did not FAIL'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
      ELSE
        CALL MPI_ERROR_CLASS(IERR, ERRORCLASS, ERR)
	IF (ERRORCLASS .NE. MPI_ERR_DIMS  .AND.    &
          ERRORCLASS .NE. MPI_ERR_ARG   .AND.    &
          ERRORCLASS .NE. MPI_ERR_OTHER)  THEN
          
          FAIL = FAIL + 1
	  WRITE(INFOBUF,101) TEST_DESC, ' Returned',  ERRORCLASS 
          
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          WRITE(INFOBUF,102)  &
            'Expected MPI_ERR_DIMS(',   MPI_ERR_DIMS,   &
            ')  OR  MPI_ERR_ARG(',    MPI_ERR_ARG,   &
            ')  OR  MPI_ERR_OTHER(',  MPI_ERR_OTHER, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ELSE
          PASS = PASS + 1
        END IF
      END IF

      CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)
!
!  REPORT OVERALL RESULTS
!
      CALL MPITEST_REPORT(PASS, FAIL, 0, TESTNAME)

      CALL MPI_FINALIZE(IERR)

      END
