!
!     Macro section
!
!     CAUTION.  Some pre-processors substitute everything after the
!     first space after the defined name, which could cause so
!     lines to overflow.  Keep definitions compact!
!
!     Fortran datatype for buffer (send and recv)
!     User must choose the biggest Fortran datatype
!     on the target platforms/host compiler (we do not use malloc).
!
!     For error checking, we also need to know how many bytes
!     one of these occupies.
!
!     See also MAX_BUFF_SIZE in mpitestf.h and data sizes in
!     mpitest_cfgf.h
!
#define MPITEST_BUF_TYPE DOUBLE COMPLEX
#define MPITEST_BUF_EXTENT 16

!     Integer which will hold an address.
#define MPITEST_AINT INTEGER

!     Comment out any of the following defines if each feature
!     is not supported by the host language compiler
!

!     Fortran command line options handling.
!     If iargc() or getargv() are not available in the host
!     compiler, comment this following MACRO out so that the
!     command line option processing will be disabled for the
!     node executable.  The library is easily modified if other
!     means are available to obtain arguments.
!     NOTE: all ranks must be passed the arguments, or customization is
!     required.

! define MPITEST_FCMDLINE 1

!     Fortran optional datatypes
!     REAL*2, *4, *8, DOUBLE COMPLEX, and INTEGER*1, *2, *4
! #define MPITEST_FREAL2_DEF         2
#define MPITEST_FREAL4_DEF          3
#define MPITEST_FREAL8_DEF          4
#define MPITEST_FDOUBLE_COMPLEX_DEF 5
#define MPITEST_FINTEGER1_DEF       6
#define MPITEST_FINTEGER2_DEF       7
#define MPITEST_FINTEGER4_DEF       8

!
!     Fortran display FORMAT for various datatype
!
! 10-31-02: I10 isn't large enough for a negative 32-bit integer.
! This shows up in the address tests when the address has the high
! bit set.
#define INT_FMT I12
#define REAL_FMT F10.2
#define COMPLEX_FMT F10.2, F10.2
#define DOUBLE_FMT D20.4
#define CHAR_FMT A
#define LOG_FMT  L1

!    Fortran optional supported datatypes
#define INT1_FMT I3
#define INT2_FMT I5
#define INT4_FMT I10

#define REAL2_FMT F10.2
#define REAL4_FMT F10.2
#define REAL8_FMT F10.2

#define DCOMPLEX_FMT D20.4, D20.4

!  This needs to be set up for different platform/host language
#define MPITEST_MAX_INT 2147483647

!  This is the minimum MPI_TAG_UB value required by the MPI Standard
#define MPITEST_TAG_UB_MIN 32767

! Compiler directives:  A few tests that don't use the common library
! (mostly derived datatypes) use these.

! Comment the following #define out to turn off receiving buffer
#define MPITEST_BUFFER_RECV   1

! Comment the following #define out to turn off recv buffer checking
! MPITEST_BUFFER_CHK should not be defined if MPITEST_BUFFER_RECV is
! NOT defined
#define MPITEST_BUFFER_CHK    2

! Comment out the following MACRO to turn off displaying
! the entire sneder & receiver buffer when error occurs
! (output file could be huge! Test may takes a long time to run
! but it is useful for debugging test failures) Off by default.
! #define MPITEST_DISP_BUF      3

! Comment the following #define out to turn off verification for
! the status object returned from MPI_Recv()
#define MPITEST_STATUS_CHK    4

! Determine whether all node should be synchronized after each
! test is done, comment it out if not desired.
#define MPITEST_SYNC          5

! Comment out the following MACRO if user wish to see all
! errors in received data buffer, otherwise, only the first
! error in each transmission will be displayed
#define MPITEST_1ST_ERR       6

